# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/CLI_DataSource.ipynb (unless otherwise specified).

__all__ = ["logger"]

# Cell

from typing import *

# Internal Cell

import os

import typer
from typer import echo
from tabulate import tabulate
import datetime as dt
import pandas as pd

from ..client import Client
from . import helper
from ..logger import get_logger, set_level

# Internal Cell

app = typer.Typer()

# Cell

logger = get_logger(__name__)

# Internal Cell


@app.command()
@helper.requires_auth_token
def s3(
    uri: str = typer.Argument(
        ..., help="The AWS S3 bucket location of the Parquet files as a string."
    ),
    access_key: Optional[str] = typer.Option(
        None,
        help="The access key for the S3 bucket. If None (default value), then the value of environment variable AWS_ACCESS_KEY_ID is used.",
    ),
    secret_key: Optional[str] = typer.Option(
        None,
        help="The secret key for the S3 bucket. If None (default value), then the value of environment variable AWS_SECRET_ACCESS_KEY is used.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output data id only.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to info and output everything",
    ),
):
    """Pull the data source from a S3 bucket to airt server."""

    from ..client import DataSource

    ds = DataSource.s3(uri=uri, access_key=access_key, secret_key=secret_key)

    progress = ds.pull()

    if quiet:
        progress.wait()

        typer.echo(f"{ds.id}")
    else:
        typer.echo(f"Pulling data ID: {ds.id}")

        progress.progress_bar()


# Internal Cell


@app.command()
@helper.requires_auth_token
def db(
    host: str = typer.Option(..., help="The name of the remote database host machine."),
    database: str = typer.Option(
        ..., help="The logical name of the database to establish the connection."
    ),
    table: str = typer.Option(..., help="The name of the table in the database."),
    port: Optional[int] = typer.Option(
        3306,
        help="The port for the database server. If the value is not passed then the default port number will be used (e.g. for MySQL, 3306 will be used)",
    ),
    engine: str = typer.Option(
        "mysql",
        help="""The name of the database engine. If the value is not passed then the
            default database engine for MySQL will be used.""",
    ),
    username: Optional[str] = typer.Option(
        None,
        "--username",
        "-u",
        help="A valid database user name. If not set (default value root), it will try to use the value from environment variable AIRT_CLIENT_DB_USERNAME.",
    ),
    password: Optional[str] = typer.Option(
        None,
        "--password",
        "-p",
        help="The password for the specified user. If not set (default value ), it will try to use the value from environment variable AIRT_CLIENT_DB_PASSWORD.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output data id only.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to info and output everything",
    ),
):
    """Creates an object that encapsulates the data from a database."""

    from ..client import DataSource

    ds = DataSource.db(
        host=host,
        database=database,
        port=port,
        table=table,
        engine=engine,
        username=username,
        password=password,
    )

    progress = ds.pull()

    if quiet:
        progress.wait()
        typer.echo(f"{ds.id}")
    else:
        typer.echo(f"Pulling data ID: {ds.id}")
        progress.progress_bar()


# Internal Cell


@app.command()
@helper.requires_auth_token
def ls(
    offset: Optional[int] = typer.Option(
        0,
        "--offset",
        "-o",
        help="The number of records to offset at the beginning of the datasource list. If None, then the default value 0 will be used.",
    ),
    limit: Optional[int] = typer.Option(
        100,
        "--limit",
        "-l",
        help="The maximum number of records to return from the server. If None, then the default value 100 will be used.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output only ids of datasource separated by space",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to info and output everything",
    ),
) -> None:
    """List available data sources."""

    logger.info("airt ds ls():")
    logger.info(f" offset - {offset}")
    logger.info(f" limit - {limit}")
    logger.info(f" quiet - {quiet}")
    logger.info(f" debug - {debug}")

    from ..client import DataSource

    df = DataSource.ls(offset=offset, limit=limit)

    df["created"] = helper.humanize_date(df["created"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    if quiet:
        ids = df["id"].astype(str).to_list()
        typer.echo("\n".join(ids))
    else:
        typer.echo(tabulate(df, headers="keys", tablefmt="plain", showindex=False))


# Internal Cell


@app.command()
@helper.requires_auth_token
def dtypes(
    id: str = typer.Argument(
        ...,
        help="The data ID in the airt service.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
) -> None:
    """Return the dtype of each column for the given data source."""

    from ..client import DataSource

    ds = DataSource(data_id=int(id))
    dtypes = ds.dtypes.T.rename(columns={0: "dtype"})
    typer.echo(dtypes)


# Internal Cell


@app.command()
@helper.requires_auth_token
def head(
    id: str = typer.Argument(
        ...,
        help="The data ID in the airt service.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
) -> None:
    """Return first few records for the given data ID."""

    from ..client import DataSource

    ds = DataSource(data_id=int(id))
    typer.echo(tabulate(ds.head(), headers="keys", tablefmt="plain", showindex=False))


# Internal Cell


@app.command()
@helper.requires_auth_token
def train(
    id: str = typer.Option(
        ...,
        "--data_id",
        "-id",
        help="The data ID in the airt service.",
    ),
    client_column: str = typer.Option(
        ...,
        "--client_column",
        help="The name of the column that uniquely identifies the users/clients as string.",
    ),
    timestamp_column: Optional[str] = typer.Option(
        None,
        "--timestamp_column",
        help="Name of the timestamp_column specifying the time of an occurred event as a string.",
    ),
    target_column: str = typer.Option(
        ...,
        "--target_column",
        help="Name of the target event for which the model needs to be trained to make predictions.",
    ),
    target: str = typer.Option(
        ...,
        "--target",
        help="Name of the target event for which the model needs to be trained to make predictions.",
    ),
    predict_after: str = typer.Option(
        ...,
        "--predict_after",
        help="Time delta in hours of the expected target event.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output model id only.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
):
    """Train a ML model for the given data source."""

    from ..client import DataSource
    from datetime import timedelta

    ds = DataSource(data_id=int(id))

    model = ds.train(
        client_column=client_column,
        target_column=target_column,
        target=target,
        predict_after=timedelta(hours=int(predict_after)),
        timestamp_column=timestamp_column,
    )

    if quiet:
        model.wait()
        typer.echo(f"{model.model_id}")
    else:
        typer.secho(f"Training started for Model ID: {model.model_id}")
        model.progress_bar()


# Internal Cell


@app.command()
@helper.requires_auth_token
def details(
    id: int = typer.Argument(
        ...,
        help="The data ID in the airt service.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
) -> None:
    """Return metadata of the given data source."""

    from ..client import DataSource

    df = DataSource.details(data_id=int(id))

    df["created"] = helper.humanize_date(df["created"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    typer.echo(tabulate(df, headers="keys", tablefmt="plain", showindex=False))


# Internal Cell


@app.command()
@helper.requires_auth_token
def rm(
    id: int = typer.Argument(
        ...,
        help="The data ID in the airt service.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output the deleted data id only.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
) -> None:
    """Remove the data source from airt service."""

    from ..client import DataSource

    df = DataSource.delete(data_id=int(id))

    df["created"] = helper.humanize_date(df["created"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    if quiet:
        typer.echo(df.iloc[0]["id"])
    else:
        typer.echo(tabulate(df, headers="keys", tablefmt="plain", showindex=False))


# Internal Cell


@app.command()
@helper.requires_auth_token
def push(
    id: int = typer.Argument(
        ...,
        help="The data ID in the airt service.",
    ),
    pred_id: int = typer.Argument(
        ...,
        help="The prediction ID in the airt service.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output the pushed prediction id only.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        "-d",
        help="Enable debug mode.",
    ),
) -> None:
    """Push the prediction results into the target data source."""

    from ..client import DataSource, Prediction

    ds = DataSource(int(id))
    pred = Prediction(prediction_id=int(pred_id), datasource_id=int(id))

    response = ds.push(pred)

    if quiet:
        typer.echo(response["prediction_id"])
    else:
        typer.echo(response)
